/*
 * communication.c
 *
 *  Created on: 26/05/2015
 *      Author: Erick Castillo
 */

#include <stdio.h>
#include <stdlib.h>
#include "communication.h"
#include "radio.h"
#include "ezradio_api_lib.h"
#include "ezradio_cmd.h"
#include "em_emu.h"
#include "em_cmu.h"
#include "rtcdriver.h"
#include "ustimer.h"


/******************************************************************************
 * Local Variables
 *****************************************************************************/
static RTCDRV_TimerID_t clockInitTimerId;			// Timer used to set initial clock
static bool clockInitTimeout = false;

static RTCDRV_TimerID_t syncTxOverTimerId;			// Timer used to set sync period
static bool slotOver = false;

static RTCDRV_TimerID_t syncTxDelayTimerId;			// Timer used to delay sync packets
static bool transmitSync = false;

static RTCDRV_TimerID_t dataTxDelayTimerId;			// Timer used to delay data packets
static bool transmitData = false;

static nodeInfo nodeInfo_table[MAX_NETWORK_NODES];	// Table to store nodes' duty cycles

#if (DEBUG_RADIO)
static time_t current_time;
static struct tm *test_time;
#endif


/******************************************************************************
 * Local Prototypes
 *****************************************************************************/
static void clockInitCallback(RTCDRV_TimerID_t id, void *user);
static void syncTxOverCallback(RTCDRV_TimerID_t id, void *user);
static void syncTxDelayCallback(RTCDRV_TimerID_t id, void *user);
static void dataTxDelayCallback(RTCDRV_TimerID_t id, void *user);
static uint16_t rand_time(uint16_t min_limit, uint16_t max_limit);


/******************************************************************************
 * Callbacks for the RTC timers
 *****************************************************************************/
static void clockInitCallback(RTCDRV_TimerID_t id, void *user)
{
	(void) id;
	(void) user;

	clockInitTimeout = true;
}

static void syncTxOverCallback(RTCDRV_TimerID_t id, void *user)
{
	(void) id;
	(void) user;

	slotOver = true;
}

static void syncTxDelayCallback(RTCDRV_TimerID_t id, void *user)
{
	(void) id;
	(void) user;

	transmitSync = true;
}

static void dataTxDelayCallback(RTCDRV_TimerID_t id, void *user)
{
	(void) id;
	(void) user;

	transmitData = true;
}

/******************************************************************************
 * @brief Obtain a random value in a range
 *****************************************************************************/
static uint16_t rand_time(uint16_t min_limit, uint16_t max_limit)
{
	int32_t ran_time = 0;

	// Obtain a random time
	ran_time = rand() % (max_limit + 1);

	if (ran_time < min_limit)
	{
		ran_time = min_limit;
	}

	return (uint16_t)ran_time;
}

/******************************************************************************
 * @brief Wait in RX mode for a sync packet to initialize local clock
 *****************************************************************************/
time_t COMMUNICATION_Init_Clk_Sync(void)
{
	radioPkt_data dataRX = {0};

	printf("Waiting for clk sync\n");

	// Allocate timers for transmission delays
	if (ECODE_EMDRV_RTCDRV_OK != RTCDRV_AllocateTimer(&syncTxDelayTimerId))
	{ printf("Failed to allocate RTC 2!!"); while(1); }
	if (ECODE_EMDRV_RTCDRV_OK != RTCDRV_AllocateTimer(&syncTxOverTimerId))
		{ printf("Failed to allocate RTC 3!!"); while(1); }
	if (ECODE_EMDRV_RTCDRV_OK != RTCDRV_AllocateTimer(&dataTxDelayTimerId))
	{ printf("Failed to allocate RTC 4!!"); while(1); }

	// Set a timeout in case there is no other nodes
	if (ECODE_EMDRV_RTCDRV_OK != RTCDRV_AllocateTimer(&clockInitTimerId)) return 0;
	if (ECODE_EMDRV_RTCDRV_OK != RTCDRV_StartTimer(clockInitTimerId, rtcdrvTimerTypeOneshot,
			CLK_INIT_TIMEOUT, clockInitCallback, NULL)) return 0;

	// Start listening
	dataRX.length = SYNC_PAYLOAD_SIZE;
	EZRADIO_StartRx(0, dataRX.length);

	// Wait for a SYNC packet or until timeout
	do {
		EMU_EnterEM2(true);

		if (radioIrqReceived)
		{
			radioIrqReceived = false;

			// Check if packet is received and read out
			if (EZRADIO_CheckReceived(&dataRX))
			{
				if (dataRX.type == SYNC_DATA)
				{
					RTCDRV_StopTimer(clockInitTimerId);
					RTCDRV_FreeTimer(clockInitTimerId);
					#if (DEBUG_RADIO)
						test_time = localtime((time_t const *) &dataRX.payload.cur_time);
						printf("%02d:%02d:%02d - init clk RX\n", test_time->tm_hour, test_time->tm_min, test_time->tm_sec);
					#endif

					// Add delay in computation time
					return dataRX.payload.cur_time;
				}
			}

			// Check if packet is received with CRC error
			if (EZRADIO_CheckCRCError())
			{
				printf("Pkt rxd - CRC Error\n");
			}
		}
	} while (!clockInitTimeout);

	RTCDRV_StopTimer(clockInitTimerId);
	RTCDRV_FreeTimer(clockInitTimerId);

	return 0;
}

/******************************************************************************
 * @brief Listen for a SYNC packet from a parent node and retransmit
 *****************************************************************************/
void COMMUNICATION_Sync_TRx(void)
{
	radioPkt_data dataRX = {0};
	radioPkt_data dataTX = {0};
	transmitSync = false;
	slotOver = false;

	// Start a timer to signal when SYNC slot is over
	RTCDRV_StartTimer(syncTxOverTimerId, rtcdrvTimerTypeOneshot, SYNC_SLOT_PERIOD, syncTxOverCallback, NULL);

	// Since there is no routing information, a static implementation between node 1 and node 2 is used,
	// node 1 transmits the clock information and node 2 updates its clock and retransmits the information
#if (NODE_ID == 0x01)

	// Delay the transmission of the SYNC packet to account for clock drift
	RTCDRV_StartTimer(syncTxDelayTimerId, rtcdrvTimerTypeOneshot, MIN_TX_DELAY_TIME, syncTxDelayCallback, NULL);

	// Stay in this loop until SYNC slot is over
	do {
		EMU_EnterEM2(true);

		if (radioIrqReceived)
		{
			radioIrqReceived = false;

			// Check if packet is sent
			if (EZRADIO_CheckTransmitted())
			{
				clearLED(LED_TRANSMIT);

				#if (DEBUG_RADIO)
					current_time = getTime();
					test_time = localtime((time_t const *) &current_time);
					printf("%02d:%02d:%02d - SYNC TX to: %02d\n", test_time->tm_hour, test_time->tm_min, test_time->tm_sec, dataTX.dest_id);
					USTIMER_Init();
					USTIMER_Delay(50u);
					USTIMER_DeInit();
				#endif
			}
		}

		// Transmit packet to all level 2 nodes in routing table
		if (transmitSync)
		{
			transmitSync = false;

			dataTX.dest_id = 0x02;
			dataTX.src_id = NODE_ID;
			dataTX.type = SYNC_DATA;
			dataTX.length = SYNC_PAYLOAD_SIZE;
			dataTX.payload.cur_time = getTime();
			EZRADIO_StartTx(0, &dataTX);
			setLED(LED_TRANSMIT);
		}
	} while (!slotOver);

	// Stop timer
	RTCDRV_StopTimer(syncTxDelayTimerId);

#else

	// Start listening for packets with N bytes
	dataRX.length = SYNC_PAYLOAD_SIZE;
	EZRADIO_StartRx(0, dataRX.length);

	// Stay in this loop until SYNC slot is over
	do {
		EMU_EnterEM2(true);

		// Check if radio interrupt is received
		if (radioIrqReceived)
		{
			radioIrqReceived = false;

			// Check if packet is sent
			if (EZRADIO_CheckTransmitted())
			{
				clearLED(LED_TRANSMIT);

				#if (DEBUG_RADIO)
					printf("SYNC TX to: %02d\n", dataTX.dest_id);
					USTIMER_Init();
					USTIMER_Delay(50u);
					USTIMER_DeInit();
				#endif
			}

			// Check if packet is received
			if (EZRADIO_CheckReceived(&dataRX))
			{
				// If SYNC packet is received, synchronize clock and retransmit
				if (dataRX.type == SYNC_DATA)
				{
					// Transmit packet to all level 3 nodes in routing table
					if (!transmitSync)
					{
						dataTX.dest_id = 0x03;
						dataTX.src_id = NODE_ID;
						dataTX.type = SYNC_DATA;
						dataTX.length = SYNC_PAYLOAD_SIZE;
						dataTX.payload.cur_time = dataRX.payload.cur_time;
						EZRADIO_StartTx(0, &dataTX);
						setLED(LED_TRANSMIT);
						transmitSync = true;
					}

					// Synchronize clock
					stopClock();
					setTime(dataRX.payload.cur_time, MIN_TX_DELAY_TIME + CLOCK_SYNC_DELAY);
					startClock();

					#if (DEBUG_RADIO)
						test_time = localtime((time_t const *) &dataRX.payload.cur_time);
						printf("%02d:%02d:%02d - SYNC RX from: %02d -> ", test_time->tm_hour, test_time->tm_min, test_time->tm_sec, dataRX.src_id);
						USTIMER_Init();
						USTIMER_Delay(50u);
						USTIMER_DeInit();
					#endif
				}
			}
		}
	} while (!slotOver);

#endif

	// Stop timer
	RTCDRV_StopTimer(syncTxOverTimerId);

	// Put the radio in sleep state until next TX/RX slot
	ezradio_change_state(EZRADIO_CMD_CHANGE_STATE_ARG_NEXT_STATE1_NEW_STATE_ENUM_SLEEP);
}

/******************************************************************************
 * @brief Transmit a data packet over RF and listen for data packets
 *****************************************************************************/
void COMMUNICATION_Data_TRx(sensorData *sensor_data, uint8_t dc)
{
	uint16_t max_tx_data_delay = 0;
	uint16_t tx_data_delay = 0;
	radioPkt_data dataRX = {0};
	radioPkt_data dataTX = {0};

	// Obtain minimum duty cycle of neighboring nodes
	max_tx_data_delay = dc;

	for (int i = 0; i < MAX_NETWORK_NODES; i++)
	{
		if (nodeInfo_table[i].node == 0) {break;}

		if (nodeInfo_table[i].dc < max_tx_data_delay)
		{
			max_tx_data_delay = nodeInfo_table[i].dc;
		}
	}

	// Convert duty cycle to time in ms
	max_tx_data_delay = ((max_tx_data_delay * PARAM_DC_PERIOD) / 100) * 1000 - SYNC_SLOT_PERIOD - MAX_SLOT_CLK_DRIFT;

	// Obtain a random time for the TX delay
	tx_data_delay = rand_time(MIN_TX_DELAY_TIME, max_tx_data_delay);

	// Start a timer to wait a random time before TX
	RTCDRV_StartTimer(dataTxDelayTimerId, rtcdrvTimerTypeOneshot, tx_data_delay, dataTxDelayCallback, NULL);

	// Start listening for packets with N bytes
	dataRX.length = NODE_PAYLOAD_SIZE;
	EZRADIO_StartRx(0, dataRX.length);

	// Stay in this loop until data slot is over
	do {
		EMU_EnterEM2(true);

		// Transmit after a random delay
		if (transmitData)
		{
			transmitData = false;

			// Stop timer
			RTCDRV_StopTimer(dataTxDelayTimerId);

			// Send DATA packet
			dataTX.dest_id = 0x02;
			dataTX.src_id = NODE_ID;
			dataTX.type = NODE_DATA;
			dataTX.length = NODE_PAYLOAD_SIZE;
			dataTX.payload.node_data.sensor_data.rhData = sensor_data->rhData;
			dataTX.payload.node_data.sensor_data.tempData = sensor_data->tempData;
			dataTX.payload.node_data.dc = dc;
			EZRADIO_StartTx(0, &dataTX);
			setLED(LED_TRANSMIT);
		}

		// Check if radio interrupt is received
		if (radioIrqReceived)
		{
			radioIrqReceived = false;

			// Check if packet is sent
			if (EZRADIO_CheckTransmitted())
			{
				clearLED(LED_TRANSMIT);

				#if (DEBUG_RADIO)
					current_time = getTime();
					test_time = localtime((time_t const *) &current_time);
					printf("%02d:%02d:%02d - DATA TX to: %02d\n", test_time->tm_hour, test_time->tm_min, test_time->tm_sec, dataTX.dest_id);
				    USTIMER_Init();
				    USTIMER_Delay(50u);
				    USTIMER_DeInit();
				#endif
			}

			// Check if packet is received
			if (EZRADIO_CheckReceived(&dataRX))
			{
				if (dataRX.type == NODE_DATA)
				{
					#if (DEBUG_RADIO)
						current_time = getTime();
						test_time = localtime((time_t const *) &current_time);
						printf("%02d:%02d:%02d - DATA RX from: %02d ", test_time->tm_hour, test_time->tm_min, test_time->tm_sec, dataRX.src_id);
						printf("-> (%05d, %05d)\n", (int)dataRX.payload.node_data.sensor_data.rhData, (int)dataRX.payload.node_data.sensor_data.tempData);
						USTIMER_Init();
						USTIMER_Delay(50u);
						USTIMER_DeInit();
					#endif

					// If new DC is received from another node, update in table
					uint8_t nodeInfo_next = 0;

					// Save duty cycles of other nodes
					for (nodeInfo_next = 0; nodeInfo_next < MAX_NETWORK_NODES; nodeInfo_next++)
					{
						if (nodeInfo_table[nodeInfo_next].node == 0 || nodeInfo_table[nodeInfo_next].node == dataRX.src_id)
						{
							break;
						}
					}

					if (nodeInfo_next < MAX_NETWORK_NODES)
					{
						nodeInfo_table[nodeInfo_next].node = dataRX.src_id;
						nodeInfo_table[nodeInfo_next].dc = dataRX.payload.node_data.dc;
					}
				}
			}

			// Check if packet is received with CRC error
			if (EZRADIO_CheckCRCError())
			{
				printf("Pkt rxd - CRC Error\n");
			}
		}
	} while (activeMode);
}
