/*
 * energy_harvesting.c
 *
 *  Created on: 06/06/2015
 *      Author: Erick Castillo
 */

#include "energy_harvesting.h"
#include "em_emu.h"
#include "em_adc.h"


/******************************************************************************
 * Local Variables
 *****************************************************************************/
// Flag used to indicate ADC is finished
static volatile bool adcConversionComplete = false;

#if (USE_TEST_VALUES)
// Used to obtain battery energy during testing
static float_t batt_energy = (INIT_BATTERY_LEVEL * BATT_EFFECTIVE_ENERGY);
static uint8_t ph_index, ph_cnt = 0;

// Solar harvested power (mW, 10 mins interval) for testing
const uint16_t power_harvested[144] = {
		1,0,2,0,0,0,3,0,0,0,0,0,0,0,1,0,0,0,3,0,0,0,0,0,
		0,3,0,0,2,0,1,1,1,2,3,3,5,9,80,80,90,90,90,95,95,
		130,130,135,135,140,140,90,145,145,145,150,60,150,
		155,155,110,160,165,140,150,130,150,145,150,120,160,
		155,150,155,135,145,140,100,135,135,135,135,130,130,
		10,15,10,15,30,35,5,8,10,10,12,14,3,3,3,4,4,4,5,5,5,
		5,5,5,5,6,6,10,10,12,12,12,12,12,5,5,5,5,6,6,7,7,8,
		8,9,9,10,10,2,1,9,2,8,2,2,0,0,1,0,0};
#endif


#if (USE_TEST_VALUES)
/******************************************************************************
 * @brief Calculate the battery's energy for testing purposes
 *****************************************************************************/
static void getTestBattEnergy(float_t dc)
{
	float_t P_harv = 0.0;
	float_t P_cons = 0.0;
	float_t E_tot = 0.0;

	// Obtain the harvested power from experimental values [mW]
	P_harv = (float_t)power_harvested[ph_index];
	ph_cnt++;
	if (ph_cnt >= 10) {
		ph_cnt = 0;
		ph_index++;
		if (ph_index >= 144) ph_index = 0;
	}

	// Adjust the harvested power with the charge efficiency of the battery
	P_harv *= BATT_CHARGE_EFFICIENCY;

	// Obtain the power consumption according to last duty cycle [mW]
	P_cons = dc*NODE_ACTIVE_PC + (1-dc)*NODE_SLEEP_PC;

	// Total energy gained (+) or consumed (-) during last minute [J]
	E_tot = ((P_harv - P_cons) * 60) / 1000;

	// New battery's energy level
	batt_energy += E_tot;

	// Check boundaries of battery's energy level
	if (batt_energy > BATT_EFFECTIVE_ENERGY) batt_energy = BATT_EFFECTIVE_ENERGY;
	else if (batt_energy < 0) batt_energy = 0;
}
#endif


/******************************************************************************
 * @brief ADC Interrupt handler (ADC0)
 *****************************************************************************/
void ADC0_IRQHandler(void)
{
   uint32_t flags;

   // Clear interrupt flags
   flags = ADC_IntGet(ADC0);
   ADC_IntClear(ADC0, flags);

   adcConversionComplete = true;
}


/******************************************************************************
 * @brief Function to measure the current battery level
 *****************************************************************************/
float_t ENERGY_HARVESTING_getBatteryLevel(uint8_t duty_cycle)
{
	uint32_t vBatt = 0;
	float_t lBatt = 0.0;

#if (USE_TEST_VALUES)
	// Use experimental values
	getTestBattEnergy((float_t)duty_cycle/100);

	// Convert battery energy to percentage level
	lBatt = batt_energy / BATT_EFFECTIVE_ENERGY;
#else
	// Sample the Vdd/3 to obtain the battery voltage
	adcConversionComplete = false;
	ADC_Start(ADC0, adcStartSingle);
	while (!adcConversionComplete) EMU_EnterEM1();
	vBatt = ADC_DataSingleGet(ADC0);

	// Convert battery voltage to percentage level
	lBatt = (float_t)(vBatt - MIN_BATTERY_LEVEL) / (MAX_BATTERY_LEVEL - MIN_BATTERY_LEVEL);
#endif

	return lBatt;
}
