/*
 * i2cim.c
 *
 *  Created on: 08/05/2015
 *      Author: Erick Castillo
 *
 * @brief I2C simple interrupt driven master mode driver for the DK/STK.
 * 		  Based on i2cspm.c
 */

#include <stddef.h>
#include "em_assert.h"
#include "em_cmu.h"
#include "em_emu.h"
#include "em_int.h"
#include "i2cidm.h"


/*******************************************************************************
 **************************   GLOBAL PARAMETERS   ******************************
 ******************************************************************************/
static I2C_TransferReturn_TypeDef status;
static CMU_Clock_TypeDef i2cClock;

/*******************************************************************************
 **************************   GLOBAL FUNCTIONS   *******************************
 ******************************************************************************/

/***************************************************************************//**
 * @brief
 *   Interrupt handler to continue transfer of data.
 *
 ******************************************************************************/
void I2C0_IRQHandler(void)
{
	status = I2C_Transfer(I2C0);
}

void I2C1_IRQHandler(void)
{
	status = I2C_Transfer(I2C1);
}


/***************************************************************************//**
 * @brief
 *   Initialize I2C peripheral
 *
 * @details
 *   This driver supports master mode only, single bus-master. In addition
 *   to configuring the I2C peripheral module, it also configures DK/STK
 *   specific setup in order to use the I2C bus.
 *
 * @param[in] init
 *   Pointer to I2C initialization structure
 ******************************************************************************/
void I2CIDM_Init(I2CSPM_Init_TypeDef *init)
{
  int i;
  I2C_Init_TypeDef i2cInit;

  EFM_ASSERT(init != NULL);

  CMU_ClockEnable(cmuClock_HFPER, true);

  /* Select I2C peripheral clock */
#if defined(I2C0)
  if (init->port == I2C0)
  {
    i2cClock = cmuClock_I2C0;
  }
#elif defined(I2C1)
  if (init->port == I2C1)
  {
    i2cClock = cmuClock_I2C1;
  }
#else
  {
    /* I2C clock is not defined */
    EFM_ASSERT(false);
    return;
  }
#endif

  CMU_ClockEnable(i2cClock, true);

  /* Output value must be set to 1 to not drive lines low. Set
     SCL first, to ensure it is high before changing SDA. */
  GPIO_PinModeSet(init->sclPort, init->sclPin, gpioModeWiredAndPullUp, 1);
  GPIO_PinModeSet(init->sdaPort, init->sdaPin, gpioModeWiredAndPullUp, 1);

  /* In some situations, after a reset during an I2C transfer, the slave
     device may be left in an unknown state. Send 9 clock pulses to
     set slave in a defined state. */
  for (i = 0; i < 9; i++)
  {
    GPIO_PinOutSet(init->sclPort, init->sclPin);
    GPIO_PinOutClear(init->sclPort, init->sclPin);
  }

  /* Enable pins and set location */
  init->port->ROUTE = I2C_ROUTE_SDAPEN |
                      I2C_ROUTE_SCLPEN |
                      (init->portLocation << _I2C_ROUTE_LOCATION_SHIFT);

  /* Set emlib init parameters */
  i2cInit.enable = true;
  i2cInit.master = true; /* master mode only */
  i2cInit.freq = init->i2cMaxFreq;
  i2cInit.refFreq = init->i2cRefFreq;
  i2cInit.clhr = init->i2cClhr;

  I2C_Init(init->port, &i2cInit);

  // Enable I2C interrupts in NVIC
#if defined(I2C0)
  NVIC_ClearPendingIRQ(I2C0_IRQn);
  NVIC_EnableIRQ(I2C0_IRQn);
#elif defined(I2C1)
  NVIC_ClearPendingIRQ(I2C1_IRQn);
  NVIC_EnableIRQ(I2C1_IRQn);
#endif
}


/***************************************************************************//**
 * @brief
 *   Perform I2C transfer
 *
 * @details
 *   This driver only supports master mode, single bus-master. It does not
 *   return until the transfer is complete, waiting on low energy mode
 *   for completion.
 *
 * @param[in] i2c
 *   Pointer to the peripheral port
 *
 * @param[in] seq
 *   Pointer to sequence structure defining the I2C transfer to take place. The
 *   referenced structure must exist until the transfer has fully completed.
 ******************************************************************************/
I2C_TransferReturn_TypeDef I2CIDM_Transfer(I2C_TypeDef *i2c, I2C_TransferSeq_TypeDef *seq)
{
  uint32_t timeout = I2CSPM_TRANSFER_TIMEOUT;

  // Use an interrupt driven transfer
  status = I2C_TransferInit(i2c, seq);

  while (status == i2cTransferInProgress && timeout--)
  {
	EMU_EnterEM1();
  }

  return status;
}
