/**************************************************************************//**
 * @file
 * @brief Energy Harvesting Sensor Node
 * @author Erick Castillo
 * @version 3.20.2
 ******************************************************************************
 * @section License
 * <b>(C) Copyright 2014 Silicon Labs, http://www.silabs.com</b>
 *******************************************************************************
 *
 * This file is licensed under the Silicon Labs Software License Agreement. See 
 * "http://developer.silabs.com/legal/version/v11/Silicon_Labs_Software_License_Agreement.txt"  
 * for details. Before using this software for any purpose, you must agree to the 
 * terms of that agreement.
 *
 ******************************************************************************/

#include <stdio.h>
#include <stdint.h>
#include <stdbool.h>
#include <stdlib.h>
#include <time.h>
#include "em_device.h"
#include "em_chip.h"
#include "em_rtc.h"
#include "em_cmu.h"
#include "em_emu.h"
#include "em_gpio.h"
#include "em_adc.h"
#include "gpiointerrupt.h"
#include "rtcdriver.h"
#include "i2cidm.h"
#include "si7013.h"
#include "ezradio_cmd.h"
#include "ezradio_api_lib.h"

#include "app_config.h"
#include "energy_harvesting.h"
#include "duty_cycling.h"
#include "communication.h"
#include "radio.h"
#if (DEBUG_DISPLAY)
#include "display.h"
#include "textdisplay.h"
#include "retargettextdisplay.h"
#elif (DEBUG_USART)
#include "retargetserial.h"
#endif
#if (DEBUG_SYNC)
#include "ustimer.h"
#endif


/******************************************************************************
 *  Local Definitions
 *****************************************************************************/
#define CLOCK_UPDATE_MS		1000			// Time between clock updates (1 sec)


/******************************************************************************
 *  Local Variables
 *****************************************************************************/
// Timer used to update main clock
RTCDRV_TimerID_t clockUpdateTimerId;

// RTC callback parameters
static void (*rtcCallback)(void*)	= 0;
static void *rtcCallbackArg			= 0;

// Manage the time slots
static volatile time_t cur_time 	= CLOCK_INIT_VALUE;	// The current time reference
static volatile uint8_t cur_secs	= 0;
static volatile uint8_t cur_mins	= 0;
static volatile bool syncUpdate		= false;			// Indicates when SYNC is needed
static uint8_t active_period 		= 1;				// Time of active period in secs
volatile bool activeMode 			= true;				// Indicates if node is in active mode

// Indicates if there is sensor data that hasn't been transmitted
static bool saved_data = false;


/******************************************************************************
 *  Local Prototypes
 *****************************************************************************/
static void clockUpdateCallback(RTCDRV_TimerID_t id, void *user);
static void gpioSetup(void);
static void adcInit(void);
static void aggregateData(uint32_t rh, int32_t temp, sensorData *sdata);


/******************************************************************************
 * @brief Function to enable SWO for energy trace
 *****************************************************************************/
#if (DEBUG_ENERGY)
void BSP_TraceSwoSetup(void)
{
  /* Enable GPIO Clock. */
  CMU->HFPERCLKEN0 |= CMU_HFPERCLKEN0_GPIO;

  /* Enable Serial wire output pin */
  GPIO->ROUTE |= GPIO_ROUTE_SWOPEN;

#if defined(_EFM32_GIANT_FAMILY) || defined(_EFM32_WONDER_FAMILY) || defined(_EFM32_LEOPARD_FAMILY)
  /* Set location 0 */
  GPIO->ROUTE = (GPIO->ROUTE & ~(_GPIO_ROUTE_SWLOCATION_MASK)) | GPIO_ROUTE_SWLOCATION_LOC0;

  /* Enable output on pin - GPIO Port F, Pin 2 */
  GPIO->P[5].MODEL &= ~(_GPIO_P_MODEL_MODE2_MASK);
  GPIO->P[5].MODEL |= GPIO_P_MODEL_MODE2_PUSHPULL;
#else
  /* Set location 1 */
  GPIO->ROUTE = (GPIO->ROUTE & ~(_GPIO_ROUTE_SWLOCATION_MASK)) | GPIO_ROUTE_SWLOCATION_LOC1;
  /* Enable output on pin */
  GPIO->P[2].MODEH &= ~(_GPIO_P_MODEH_MODE15_MASK);
  GPIO->P[2].MODEH |= GPIO_P_MODEH_MODE15_PUSHPULL;
#endif

  /* Enable debug clock AUXHFRCO */
  CMU->OSCENCMD = CMU_OSCENCMD_AUXHFRCOEN;

  /* Wait until clock is ready */
  while(!(CMU->STATUS & CMU_STATUS_AUXHFRCORDY));

  /* Enable trace in core debug */
  CoreDebug->DHCSR |= 1;
  CoreDebug->DEMCR |= CoreDebug_DEMCR_TRCENA_Msk;

  /* Enable PC and IRQ sampling output */
  DWT->CTRL = 0x400113FF;
  /* Set TPIU prescaler to 16. */
  TPI->ACPR = 0xf;
  /* Set protocol to NRZ */
  TPI->SPPR = 2;
  /* Disable continuous formatting */
  TPI->FFCR = 0x100;
  /* Unlock ITM and output data */
  ITM->LAR = 0xC5ACCE55;
  ITM->TCR = 0x10009;
}
#endif


/******************************************************************************
 * @brief Function to test output values of LQ
 *****************************************************************************/
#if (DEBUG_DC)
static void test_lin_quad_tracking(uint8_t dc, float_t bat_level)
{
	struct tm *time;
	time = localtime((time_t const *) &cur_time);
	int8_t batt = bat_level*100;

	printf("%02d:%02d,%d,%d\n", time->tm_hour, time->tm_min, dc, batt);
}
#endif


/******************************************************************************
 * @brief  Main function
 *****************************************************************************/
int main(void)
{
	I2CSPM_Init_TypeDef i2cInit = I2CSPM_INIT_DEFAULT;
	uint8_t duty_cycle = APP_INIT_DC;
	float_t battery = 0.0;
	sensorData sens_data = {0};
	sensorData avgSens_data = {0};


	// Chip errata
	CHIP_Init();

	// Configure HFRC frequency
	CMU_HFRCOBandSet(cmuHFRCOBand_28MHz);

#if (DEBUG_ENERGY)
	// Trace energy profile
	BSP_TraceSwoSetup();
#endif

	// Enable GPIO for sensor readings and LEDs
	gpioSetup();

	// Enable ADC for battery voltage readings
	adcInit();

#if (DEBUG_DISPLAY)
	// Initialize the display module
	DISPLAY_Init();
	RETARGET_TextDisplayInit();
	printf("\f");
#elif (DEBUG_USART)
	// Enable UART(VCOM) for testing
	RETARGET_SerialInit();
	RETARGET_SerialCrLf(1);
	printf("\f");
#endif

	// Initialize RTC
	RTCDRV_Init();

	// Initialize sensor bus
	I2CIDM_Init(&i2cInit);

	// Initialize radio and reset FIFOs
	EZRADIO_Init();
	EZRADIO_ResetTRxFifo();

	// Set up the seed for random TX times
	srand(NODE_ID);

	// Get initial sensor status
	if (!Si7013_Detect(i2cInit.port, SI7021_ADDR, NULL))
	{ printf("Failed to detect sensor!!"); while(1); }

	// Initialize clock
	time_t network_time = COMMUNICATION_Init_Clk_Sync();
	if (network_time > 0)
	{
		setTime(network_time, CLOCK_INIT_DELAY);
	}
	printf("Clock set\n");

	// Set up RTC
	if (ECODE_EMDRV_RTCDRV_OK != RTCDRV_AllocateTimer(&clockUpdateTimerId))
	{ printf("Failed to allocate RTC!!"); while(1); }
	if (ECODE_EMDRV_RTCDRV_OK != RTCDRV_StartTimer(clockUpdateTimerId, rtcdrvTimerTypePeriodic,
			CLOCK_UPDATE_MS, clockUpdateCallback, NULL))
	{ printf("Failed to start RTC!!"); while(1); }


	// Infinite loop
	while (1)
	{
		// Node wakes up
		setLED(LED_ACTIVE);

		// Measure battery level
		battery = ENERGY_HARVESTING_getBatteryLevel(duty_cycle);

		// Check if there is enough battery to operate
		if (battery > 0)
		{
			// Update duty cycle
			duty_cycle = DUTY_CYCLING_GetDC(battery);
			active_period = (duty_cycle * PARAM_DC_PERIOD) / 100;

			// Turn the radio on
			EZRADIO_Restart();

			// Enter SYNC slot after each SYNC period
			if (syncUpdate)
			{
				syncUpdate = false;

				COMMUNICATION_Sync_TRx();
			}

			// Get sensor readings
			Si7013_MeasureRHAndTemp(i2cInit.port, SI7021_ADDR, &sens_data.rhData, &sens_data.tempData);

			// If duty cycle is very low aggregate data, only transmit when it's above a threshold
			if (duty_cycle < APP_MIN_DC_TX)
			{
				aggregateData(sens_data.rhData, sens_data.tempData, &avgSens_data);
				#if (DEBUG_RADIO)
					printf("NO ENERGY FOR TX - DATA AGGREGATED\n");
				#endif

				active_period = 0;
			}
			else
			{
				if (saved_data)
				{
					aggregateData(sens_data.rhData, sens_data.tempData, &avgSens_data);
					sens_data.rhData = avgSens_data.rhData;
					sens_data.tempData = avgSens_data.tempData;
					saved_data = false;
				}

				// Transmit node's data
				COMMUNICATION_Data_TRx(&sens_data, duty_cycle);
			}

			// Turn the radio off
			EZRADIO_Shutdown();
		}
		else
		{
			duty_cycle = 0;
			active_period = 0;
			EZRADIO_Shutdown();
		}

		#if (DEBUG_DC)
			test_lin_quad_tracking(duty_cycle, battery);
		#endif

		// Node goes to sleep
		clearLED(LED_ACTIVE);
		do {
			EMU_EnterEM2(true);
		} while (!activeMode);
	}
}


/******************************************************************************
 * @brief   Register a callback function at the given frequency.
 *
 * @param[in] pFunction  Pointer to function that should be called at the
 *                       given frequency.
 * @param[in] argument   Argument to be given to the function.
 * @param[in] frequency  Frequency at which to call function at.
 *
 * @return  0 for successful or
 *         -1 if the requested frequency does not match the RTC frequency.
 *****************************************************************************/
int RtcIntCallbackRegister(void (*pFunction)(void*),
                           void* argument,
                           unsigned int frequency)
{
  rtcCallback    = pFunction;
  rtcCallbackArg = argument;

  return 0;
}


/******************************************************************************
 * @brief Callback used to count between measurement updates
 *****************************************************************************/
static void clockUpdateCallback(RTCDRV_TimerID_t id, void *user)
{
	(void) id;
	(void) user;

	cur_time++;
	cur_secs++;

	// Enter active mode when DC period ends
	if (cur_secs >= PARAM_DC_PERIOD)
	{
		activeMode = true;
		cur_secs -= PARAM_DC_PERIOD;
		cur_mins++;
	}
	// Active mode ends
	else if (cur_secs >= active_period)
	{
		activeMode = false;
	}

	// Change slot
	if (cur_mins >= PARAM_SYNC_PERIOD)
	{
		cur_mins -= PARAM_SYNC_PERIOD;
		syncUpdate = true;
	}

#if (DEBUG_SYNC)
	if (activeMode)
	{
		struct tm *time;
		time = localtime((time_t const *) &cur_time);
		printf("\f%02d:%02d", time->tm_min, time->tm_sec);
		USTIMER_Init();
		USTIMER_DelayIntSafe(50u);
		USTIMER_DeInit();
	}
#endif
}


/******************************************************************************
 * @brief Setup GPIO for sensor readings (I2C) and LEDs
 *****************************************************************************/
static void gpioSetup(void)
{
	// Enable GPIO clock
	CMU_ClockEnable(cmuClock_GPIO, true);

	// Initialize GPIO interrupt
	GPIOINT_Init();

	// Enable si7021 sensor isolation switch
	GPIO_PinModeSet(gpioPortF, 8, gpioModePushPull, 1);

	// Enable LEDs
	GPIO_PinModeSet(gpioPortF, LED_ACTIVE, gpioModePushPull, 0);
	GPIO_PinModeSet(gpioPortF, LED_TRANSMIT, gpioModePushPull, 0);
}


/**************************************************************************//**
 * @brief ADC Initialization
 *****************************************************************************/
static void adcInit(void)
{
   ADC_Init_TypeDef init = ADC_INIT_DEFAULT;
   ADC_InitSingle_TypeDef initSingle = ADC_INITSINGLE_DEFAULT;

   // Enable ADC clock
   CMU_ClockEnable(cmuClock_ADC0, true);

   // Initiate ADC peripheral
   ADC_Init(ADC0, &init);

   // Setup single conversions for internal VDD/3
   initSingle.acqTime = adcAcqTime16;
   initSingle.input = adcSingleInpVDDDiv3;
   ADC_InitSingle(ADC0, &initSingle);

   // Manually set some calibration values
   ADC0->CAL = (0x7C << _ADC_CAL_SINGLEOFFSET_SHIFT) | (0x1F << _ADC_CAL_SINGLEGAIN_SHIFT);

   // Enable interrupt on completed conversion
   ADC_IntEnable(ADC0, ADC_IEN_SINGLE);
   NVIC_ClearPendingIRQ(ADC0_IRQn);
   NVIC_EnableIRQ(ADC0_IRQn);
}


/******************************************************************************
 * @brief Aggregate data when DC is very low
 *****************************************************************************/
static void aggregateData(uint32_t rh, int32_t temp, sensorData *sdata)
{
	if (!saved_data)
	{
		sdata->rhData = rh;
		sdata->tempData = temp;
		saved_data = true;
	}
	else
	{
		sdata->rhData = (sdata->rhData + rh) / 2;
		sdata->tempData = (sdata->tempData + temp) / 2;
	}
}


/******************************************************************************
 * @brief Set a LED on
 *****************************************************************************/
void setLED(uint8_t led)
{
    GPIO_PinOutSet(gpioPortF, led);
}


/******************************************************************************
 * @brief Set a LED off
 *****************************************************************************/
void clearLED(uint8_t led)
{
    GPIO_PinOutClear(gpioPortF, led);
}


/******************************************************************************
 * @brief Get the current time in seconds
 *****************************************************************************/
time_t getTime(void)
{
	return cur_time;
}


/******************************************************************************
 * @brief Update the RTC clock
 *****************************************************************************/
void setTime(time_t new_time, uint32_t offset_ms)
{
	bool running = false;

	RTCDRV_IsRunning(clockUpdateTimerId, &running);
	if (running) { stopClock(); }

	cur_time = new_time + 1;
	cur_secs = 1;

	RTCDRV_Delay(1000-offset_ms);
}


/******************************************************************************
 * @brief Start the RTC clock
 *****************************************************************************/
void startClock(void)
{
	RTCDRV_StartTimer(clockUpdateTimerId, rtcdrvTimerTypePeriodic, CLOCK_UPDATE_MS, clockUpdateCallback, NULL);
}


/******************************************************************************
 * @brief Stop the RTC clock
 *****************************************************************************/
void stopClock(void)
{
	RTCDRV_StopTimer(clockUpdateTimerId);
}
