/*
 * radio.c
 *
 *  Created on: 15/05/2015
 *      Author: Erick Castillo
 */

#include <stddef.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>
#include "em_gpio.h"
#include "em_prs.h"
#include "em_cmu.h"
#include "gpiointerrupt.h"
#include "radio.h"
#include "ezradio_cmd.h"
#include "ezradio_prop.h"
#include "ezradio_hal.h"
#include "ezradio_api_lib.h"
#include "ezradio_api_lib_add.h"


/******************************************************************************
 * Local Variables
 *****************************************************************************/
// Radio configuration data array.
const uint8_t Radio_Configuration_Data_Array[] = RADIO_CONFIGURATION_DATA_ARRAY;

// Radio and application configuration data array.
const tRadioConfiguration RadioConfiguration = RADIO_CONFIGURATION_DATA;

// Pointer to the radio configuration data array.
const tRadioConfiguration * const pRadioConfiguration = &RadioConfiguration;

// Packet data array
uint8_t radioPkt[RADIO_MAX_PACKET_LENGTH];

// Radio interrupt receive flag
bool radioIrqReceived = false;


/******************************************************************************
 * Local Prototypes
 *****************************************************************************/
static void ezradioPowerUp(void);


/******************************************************************************
 * @brief Radio GPIO interrupt
 *****************************************************************************/
void GPIO_EZRadio_INT_IRQHandler(uint8_t pin)
{
  (void)pin;

  // Sign radio interrupt received
  radioIrqReceived = true;
}


/******************************************************************************
 * @brief Power up the radio
 *****************************************************************************/
static void ezradioPowerUp(void)
{
  volatile uint16_t wDelay = 0u;

  // Hardware reset the chip
  ezradio_reset();

  // Wait until reset timeout or Reset IT signal
  for (; wDelay < pRadioConfiguration->Radio_Delay_Cnt_After_Reset; wDelay++);
}


/******************************************************************************
 * @brief Initialize the radio
 * @author Sz. Papp
 *****************************************************************************/
void EZRADIO_Init(void)
{
  uint16_t wDelay;
  uint8_t  pktPosCnt;

  // Initialize radio GPIOs and SPI port
  ezradio_hal_GpioInit(GPIO_EZRadio_INT_IRQHandler);

  // Setup PRS for PTI pins
  CMU_ClockEnable(cmuClock_PRS, true);

  // Enable RF_GPIO0 and RF_GPIO1 pins
  GPIO_PinModeSet((GPIO_Port_TypeDef)RF_GPIO0_PORT, RF_GPIO0_PIN, gpioModeInput, 0);
  GPIO_PinModeSet((GPIO_Port_TypeDef)RF_GPIO1_PORT, RF_GPIO1_PIN, gpioModeInput, 0);

  // Pin PA0 and PA1 output the GPIO0 and GPIO1 via PRS to PTI
  GPIO_PinModeSet(gpioPortA, 0, gpioModePushPull, 0);
  GPIO_PinModeSet(gpioPortA, 1, gpioModePushPull, 0);

  // Configure INT/PRS channels
  GPIO_IntConfig((GPIO_Port_TypeDef)RF_GPIO0_PORT, RF_GPIO0_PIN,
                 false, false, false);
  GPIO_IntConfig((GPIO_Port_TypeDef)RF_GPIO1_PORT, RF_GPIO1_PIN,
                 false, false, false);

  // Setup PRS
  PRS_SourceAsyncSignalSet(0, PRS_CH_CTRL_SOURCESEL_GPIOH, PRS_CH_CTRL_SIGSEL_GPIOPIN15);
  PRS_SourceAsyncSignalSet(1, PRS_CH_CTRL_SOURCESEL_GPIOH, PRS_CH_CTRL_SIGSEL_GPIOPIN14);
  PRS->ROUTE = (PRS_ROUTE_CH0PEN | PRS_ROUTE_CH1PEN);

  // Make sure PRS sensing is enabled (should be by default)
  GPIO_InputSenseSet(GPIO_INSENSE_PRS, GPIO_INSENSE_PRS);

  // Initialize communication with the radio chip
  ezradio_hal_SpiInit();

  // Power up the radio chip
  ezradioPowerUp();

  // Load radio configuration
  while (EZRADIO_CONFIG_SUCCESS!= ezradio_configuration_init(pRadioConfiguration->Radio_ConfigurationArray))
  {
    /* Error hook */
#ifdef ERROR_HOOK
    ERROR_HOOK;
#else
    printf("ERROR: Radio configuration failed!\n");
#endif
    for (wDelay = 0x7FFF; wDelay--; ) ;

    // Try again to power up the radio chip
    ezradioPowerUp();
  }

  // Read ITs, clear pending ones
  ezradio_get_int_status(0u, 0u, 0u);

  // Initialize packet content ****check if needed
  for(pktPosCnt = 0u; pktPosCnt < pRadioConfiguration->Radio_PacketLength; pktPosCnt++)
  {
    radioPkt[pktPosCnt] = pRadioConfiguration->Radio_CustomPayload[pktPosCnt];
  }
}


/******************************************************************************
 * @brief Shuts down the radio
 *****************************************************************************/
void EZRADIO_Shutdown(void)
{
	ezradio_reset();
}


/******************************************************************************
 * @brief Reboots the radio
 *****************************************************************************/
void EZRADIO_Restart(void)
{
	uint16_t wDelay;
	uint8_t  pktPosCnt;

	ezradioPowerUp();

	//Load radio configuration
	while (EZRADIO_CONFIG_SUCCESS!= ezradio_configuration_init(pRadioConfiguration->Radio_ConfigurationArray))
	{
		// Error hook
		#ifdef ERROR_HOOK
			ERROR_HOOK;
		#else
			printf("ERROR: Radio configuration failed!\n");
		#endif
			for (wDelay = 0x7FFF; wDelay--; ) ;

		// Power Up the radio chip
		ezradioPowerUp();
	}

	// Read ITs, clear pending ones
	ezradio_get_int_status(0u, 0u, 0u);

	// Initialize packet content
	for(pktPosCnt = 0u; pktPosCnt < pRadioConfiguration->Radio_PacketLength; pktPosCnt++)
	{
		radioPkt[pktPosCnt] = pRadioConfiguration->Radio_CustomPayload[pktPosCnt];
	}
}


/******************************************************************************
 * @brief Set radio to TX mode, switch to RX when done
 * @return true if packet is issued to transmit
 *****************************************************************************/
bool EZRADIO_StartTx(uint8_t channel, radioPkt_data *data)
{
	ezradio_request_device_state();

	if (ezradioReply.REQUEST_DEVICE_STATE.CURR_STATE == EZRADIO_CMD_REQUEST_DEVICE_STATE_REP_CURR_STATE_MAIN_STATE_ENUM_TX) {
		return false;
	}

	// Build packet structure
	*(uint8_t *)(radioPkt + APP_PKT_MATCH_LOC) = data->dest_id;
	*(uint8_t *)(radioPkt + APP_PKT_LEN_LOC) = data->length;
	*(uint8_t *)(radioPkt + APP_PKT_SRCID_LOC) = data->src_id;
	*(uint8_t *)(radioPkt + APP_PKT_TYPE_LOC) = data->type;

	switch (data->type)
	{
	case NODE_DATA:
		*(uint32_t *)(radioPkt + APP_PKT_DATA_LOC) = data->payload.node_data.sensor_data.rhData;
		*(uint32_t *)(radioPkt + APP_PKT_DATA_LOC + 4) = data->payload.node_data.sensor_data.tempData;
		*(uint8_t *)(radioPkt + APP_PKT_DATA_LOC + 8) = data->payload.node_data.dc;
		break;
	case SYNC_DATA:
		*(uint32_t *)(radioPkt + APP_PKT_DATA_LOC) = data->payload.cur_time;
		break;
	default:
		break;
	}

	// Fill the TX FIFO with data
	ezradio_write_tx_fifo(data->length, (uint8_t *)radioPkt);

	// Start sending packet, START immediately, Packet n bytes long, go to RX when done ****change when want to retransmit
	ezradio_start_tx(channel, 0x80, data->length);

	return true;
}


/******************************************************************************
 * @brief Set Radio to RX mode
 *****************************************************************************/
void EZRADIO_StartRx(uint8_t channel, uint8_t pkt_length)
{
	// Start receiving packet, START immediately, Packet n bytes long
    ezradio_start_rx(channel, 0u, (uint16_t)pkt_length,
    //ezradio_start_rx(channel, 0u, 0u,
                  EZRADIO_CMD_START_RX_ARG_NEXT_STATE1_RXTIMEOUT_STATE_ENUM_NOCHANGE,
                  EZRADIO_CMD_START_RX_ARG_NEXT_STATE2_RXVALID_STATE_ENUM_RX,
                  EZRADIO_CMD_START_RX_ARG_NEXT_STATE3_RXINVALID_STATE_ENUM_RX );
}


/******************************************************************************
 * @brief Check if Packet received IT flag is pending
 * @return   true - Packet successfully received / false - No packet pending
 *****************************************************************************/
bool EZRADIO_CheckReceived(radioPkt_data *data)
{
  if (false == ezradio_hal_NirqLevel())
  {
    // Read ITs, clear pending ones
    ezradio_get_int_status(~(EZRADIO_CMD_GET_INT_STATUS_REP_PH_PEND_PACKET_RX_PEND_BIT), 0u, 0u);

    // Check the reason for the IT
    if (ezradioReply.GET_INT_STATUS.PH_PEND & EZRADIO_CMD_GET_INT_STATUS_REP_PH_PEND_PACKET_RX_PEND_BIT)
    {
      // Read the data from RX FIFO
      ezradio_read_rx_fifo(data->length, (uint8_t *) &radioPkt[0u]);

      // Read out individual fields
      //data->length = *(uint8_t *)(radioPkt + 1u);
      data->src_id = *(uint8_t *)(radioPkt + 2u);
      data->type = *(uint8_t *)(radioPkt + 3u);

      switch (data->type)
      	{
      	case NODE_DATA:
            data->payload.node_data.sensor_data.rhData = *(uint32_t *)(radioPkt + APP_PKT_DATA_LOC);
            data->payload.node_data.sensor_data.tempData = *(uint32_t *)(radioPkt + APP_PKT_DATA_LOC + 4u);
            data->payload.node_data.dc = *(uint8_t *)(radioPkt + APP_PKT_DATA_LOC + 8u);
      		break;
      	case SYNC_DATA:
            data->payload.cur_time = *(uint32_t *)(radioPkt + APP_PKT_DATA_LOC);
      		break;
      	default:
      		break;
      	}

      return true;
    }

    // Reset FIFO
    ezradio_fifo_info(EZRADIO_CMD_FIFO_INFO_ARG_FIFO_RX_BIT);
  }

  return false;
}


/******************************************************************************
 * @brief Check if Packet sent IT flag is pending
 * @return   true / false
 *****************************************************************************/
bool EZRADIO_CheckTransmitted(void)
{
  if (false == ezradio_hal_NirqLevel())
  {
    // Read ITs, clear pending ones
    ezradio_get_int_status(~(EZRADIO_CMD_GET_INT_STATUS_REP_PH_PEND_PACKET_SENT_PEND_BIT), 0u, 0u);

    // check the reason for the IT
    if (ezradioReply.GET_INT_STATUS.PH_PEND & EZRADIO_CMD_GET_INT_STATUS_REP_PH_PEND_PACKET_SENT_PEND_BIT)
    {
      return true;
    }
  }

  return false;
}


/******************************************************************************
 * @brief Check if Packet CRC error IT flag is pending
 * @return   true - Packet CRC error / false - No error
 *****************************************************************************/
bool EZRADIO_CheckCRCError(void)
{
  if (false == ezradio_hal_NirqLevel())
  {
    // Read ITs, clear pending ones
    ezradio_get_int_status(~(EZRADIO_CMD_GET_INT_STATUS_ARG_PH_CLR_PEND_CRC_ERROR_PEND_CLR_MASK), 0u, 0u);

    // Check the reason for the IT
    if (ezradioReply.GET_INT_STATUS.PH_PEND & EZRADIO_CMD_GET_INT_STATUS_REP_PH_PEND_CRC_ERROR_PEND_BIT)
    {
      // Reset FIFO
      //ezradio_change_state(EZRADIO_CMD_CHANGE_STATE_ARG_NEXT_STATE1_NEW_STATE_ENUM_READY);
      ezradio_fifo_info(EZRADIO_CMD_FIFO_INFO_ARG_FIFO_RX_BIT);
      ezradio_change_state(EZRADIO_CMD_CHANGE_STATE_ARG_NEXT_STATE1_NEW_STATE_ENUM_RX);
      return true;
    }
  }
  return false;
}


/******************************************************************************
 * @brief Resets both the TX and RX FIFOs
 *****************************************************************************/
void EZRADIO_ResetTRxFifo(void)
{
  ezradio_fifo_info(EZRADIO_CMD_FIFO_INFO_ARG_FIFO_RX_BIT | EZRADIO_CMD_FIFO_INFO_ARG_FIFO_TX_BIT);
}
